@@ -404,16 +404,16 @@ void wikitotext(TCXSTR *rbuf, const char *str){
     } else if(*line == '@'){
       line++;
       if(*line == '@') line++;
-      bool object = false;
+      bool obj = false;
       if(*line == '!'){
-        object = true;
+        obj = true;
         line++;
       }
       while(*line == '<' || *line == '>' || *line == '+' || *line == '-' || *line == '|'){
         line++;
       }
       line = tcstrskipspc(line);
-      if(*line != '\0') tcxstrprintf(rbuf, "  [%s:%s]\n\n", object ? "OBJECT" : "IMAGE", line);
+      if(*line != '\0') tcxstrprintf(rbuf, "  [%s:%s]\n\n", obj ? "OBJECT" : "IMAGE", line);
       ri++;
     } else if(tcstrfwm(line, "===")){
       tcxstrcat2(rbuf, "----\n\n");
@@ -767,14 +767,15 @@ void wikitohtml(TCXSTR *rbuf, const char *str, const char *id, const char *buri,
       ri = ei + 1;
     } else if(*line == '@'){
       line++;
-      bool anchor = false;
+      imgcnt++;
+      bool anc = false;
       if(*line == '@'){
-        anchor = true;
+        anc = true;
         line++;
       }
-      bool object = false;
+      bool obj = false;
       if(*line == '!'){
-        object = true;
+        obj = true;
         line++;
       }
       const char *align = "normal";
@@ -817,6 +818,7 @@ void wikitohtml(TCXSTR *rbuf, const char *str, const char *id, const char *buri,
         bool wratio = false;
         int height = 0;
         bool hratio = false;
+        const char *alt = NULL;
         TCLIST *params = NULL;
         char *sep = strchr(uri, '|');
         if(sep){
@@ -835,7 +837,15 @@ void wikitohtml(TCXSTR *rbuf, const char *str, const char *id, const char *buri,
             }
             if(*sep == '%') hratio = true;
             sep = strchr(sep, '|');
-            if(sep) params = tcstrsplit(sep + 1, "|");
+            if(sep){
+              sep++;
+              alt = sep;
+              sep = strchr(sep, '|');
+              if(sep){
+                *(sep++) = '\0';
+                params = tcstrsplit(sep, "|");
+              }
+            }
           }
         }
         const char *name;
@@ -852,8 +862,8 @@ void wikitohtml(TCXSTR *rbuf, const char *str, const char *id, const char *buri,
         const char *scale = width > 0 ? "sized" : "ratio";
         tcxstrprintf(rbuf, "<div class=\"image image_%s image_%s%d image_%s\">",
                      align, align, lv, scale);
-        if(anchor) tcxstrprintf(rbuf, "<a href=\"%@\">", url);
-        if(object){
+        if(anc) tcxstrprintf(rbuf, "<a href=\"%@\">", url);
+        if(obj){
           tcxstrprintf(rbuf, "<object data=\"%@\"", url);
           const char *type = mimetype(url);
           if(type) tcxstrprintf(rbuf, " type=\"%@\"", type);
@@ -867,14 +877,24 @@ void wikitohtml(TCXSTR *rbuf, const char *str, const char *id, const char *buri,
                            tclistval2(params, i), tclistval2(params, i + 1));
             }
           }
-          tcxstrprintf(rbuf, "object:%d:%@</object>", ++imgcnt, name);
+          if(alt){
+            tcxstrprintf(rbuf, "%@", alt);
+          } else {
+            tcxstrprintf(rbuf, "object:%d:%@", imgcnt, name);
+          }
+          tcxstrprintf(rbuf, "</object>");
         } else {
           tcxstrprintf(rbuf, "<img src=\"%@\"", url);
           if(width > 0) tcxstrprintf(rbuf, " width=\"%d%s\"", width, wratio ? "%" : "");
           if(height > 0) tcxstrprintf(rbuf, " height=\"%d%s\"", height, hratio ? "%" : "");
-          tcxstrprintf(rbuf, " alt=\"image:%d:%@\" />", ++imgcnt, name);
+          if(alt){
+            tcxstrprintf(rbuf, " alt=\"%@\"", alt);
+          } else {
+            tcxstrprintf(rbuf, " alt=\"image:%d:%@\"", imgcnt, name);
+          }
+          tcxstrprintf(rbuf, " />");
         }
-        if(anchor) tcxstrprintf(rbuf, "</a>");
+        if(anc) tcxstrprintf(rbuf, "</a>");
         tcxstrprintf(rbuf, "</div>\n");
         tcfree(url);
         if(params) tclistdel(params);
