@@ -364,6 +364,7 @@ static void dosession(TCMPOOL *mpool){
   const char *p_wiki = tcstrskipspc(tcmapget4(params, "wiki", ""));
   bool p_mts = *tcmapget4(params, "mts", "") != '\0';
   const char *p_hash = tcstrskipspc(tcmapget4(params, "hash", ""));
+  uint32_t p_seskey = tcatoi(tcmapget4(params, "seskey", ""));
   const char *p_comowner = tcstrskipspc(tcmapget4(params, "comowner", ""));
   const char *p_comtext = tcstrskipspc(tcmapget4(params, "comtext", ""));
   const char *p_ummode = tcstrskipspc(tcmapget4(params, "ummode", ""));
@@ -393,6 +394,8 @@ static void dosession(TCMPOOL *mpool){
   const char *p_scheme = (rp && *rp != '\0') ? "https" : "http";
   const char *p_scripturl = tcmpoolpushptr(mpool, tcsprintf("%s://%s%s",
                                                             p_scheme, p_hostname, g_scriptname));
+  rp = getenv("HTTP_REFERER");
+  const char *p_referrer = rp ? rp : "";
   rp = getenv("HTTP_USER_AGENT");
   const char *p_useragent = rp ? rp : "";
   const char *p_userlang = "";
@@ -415,6 +418,7 @@ static void dosession(TCMPOOL *mpool){
   // perform authentication
   bool auth = true;
   const char *userinfo = NULL;
+  uint32_t seskey = 0;
   const char *authcookie = NULL;
   const char *ridque = "";
   const char *ridans = "";
@@ -439,9 +443,12 @@ static void dosession(TCMPOOL *mpool){
         token[authsiz] = '\0';
         TCLIST *elems = tcmpoolpushlist(mpool, tcstrsplit(token, ":"));
         if(tclistnum(elems) >= 4 && !strcmp(tclistval2(elems, 0), salt)){
-          p_user = tclistval2(elems, 1);
-          p_pass = tclistval2(elems, 2);
-          cont = true;
+          seskey = tcatoi(tclistval2(elems, 3));
+          if(seskey > 0){
+            p_user = tclistval2(elems, 1);
+            p_pass = tclistval2(elems, 2);
+            cont = true;
+          }
         }
       }
     }
@@ -456,13 +463,27 @@ static void dosession(TCMPOOL *mpool){
         if(!strcmp(hash, numbuf)){
           auth = true;
           userinfo = pv ? pv : "";
+          if(seskey < 1){
+            uint32_t seed = 19780211;
+            for(rp = p_pass; *rp != '\0'; rp++){
+              seed = seed * 31 + *(unsigned char *)rp;
+            }
+            double integ;
+            double fract = modf(now, &integ) * (1ULL << 31);
+            seskey = (((uint32_t)integ + (uint32_t)fract) ^ (seed << 8)) & INT32_MAX;
+            if(seskey < 1) seskey = INT32_MAX;
+          }
           int tsiz = strlen(p_user) + strlen(p_pass) + saltsiz + NUMBUFSIZ * 2;
           char token[tsiz];
-          tsiz = sprintf(token, "%s:%s:%s:%lld", salt, p_user, p_pass, (long long)tctime());
+          tsiz = sprintf(token, "%s:%s:%s:%u:%lld",
+                         salt, p_user, p_pass, (unsigned int)seskey, (long long)now);
           tcmd5hash(token, tsiz, numbuf);
           sprintf(token + tsiz, ":%s", numbuf);
           tcarccipher(token, tsiz, salt, saltsiz, token);
-          if(!cont) authcookie = tcmpoolpushptr(mpool, tcurlencode(token, tsiz));
+          if(!cont){
+            authcookie = tcmpoolpushptr(mpool, tcurlencode(token, tsiz));
+            p_seskey = seskey;
+          }
         }
       }
     }
@@ -484,6 +505,7 @@ static void dosession(TCMPOOL *mpool){
     p_user = "";
     p_pass = "";
     auth = false;
+    seskey = 0;
     authcookie = "";
   }
   bool admin = auth && !strcmp(p_user, ADMINNAME);
@@ -517,6 +539,17 @@ static void dosession(TCMPOOL *mpool){
   int omode = TDBOREADER;
   if(!strcmp(p_act, "update") && auth && post) omode = TDBOWRITER;
   if(!strcmp(p_act, "comment") && cancom && post) omode = TDBOWRITER;
+  if(post && auth && *p_referrer != '\0'){
+    char *src = tcmpoolpushptr(mpool, tcstrdup(p_referrer));
+    char *wp = strchr(src, '?');
+    if(wp) *wp = '\0';
+    if(strcmp(src, p_scripturl)){
+      tclistprintf(emsgs, "Referrer is invalid (%s).", src);
+      admin = false;
+      post = false;
+      omode = TDBOREADER;
+    }
+  }
   if(!tctdbopen(tdb, g_database, omode))
     setdberrmsg(emsgs, tdb, "Opening the database was failed.");
   int64_t mtime = tctdbmtime(tdb);
@@ -676,7 +709,9 @@ static void dosession(TCMPOOL *mpool){
     }
   } else if(!strcmp(p_act, "update")){
     // update view
-    if(p_id > 0){
+    if(seskey > 0 && p_seskey != seskey){
+      tclistprintf(emsgs, "The session key is invalid (%u).", (unsigned int)p_seskey);
+    } else if(p_id > 0){
       TCMAP *cols = tcmpoolpushmap(mpool, dbgetart(tdb, p_id));
       if(cols){
         if(checkfrozen(cols) && !admin){
@@ -757,7 +792,9 @@ static void dosession(TCMPOOL *mpool){
     if(g_users){
       if(admin){
         if(post && p_umname != '\0'){
-          if(!strcmp(p_ummode, "new")){
+          if(seskey > 0 && p_seskey != seskey){
+            tclistprintf(emsgs, "The session key is invalid (%u).", (unsigned int)p_seskey);
+          } else if(!strcmp(p_ummode, "new")){
             if(tcmapget2(g_users, p_umname)){
               tclistprintf(emsgs, "The user already exists.");
             } else if(!checkusername(p_umname)){
@@ -881,7 +918,9 @@ static void dosession(TCMPOOL *mpool){
     if(g_upload && tcstatfile(g_upload, &isdir, NULL, &mtime) && isdir){
       if(auth){
         if(post && (p_fmfilebuf || *p_fmpath != '\0')){
-          if(!strcmp(p_fmmode, "new")){
+          if(seskey > 0 && p_seskey != seskey){
+            tclistprintf(emsgs, "The session key is invalid (%u).", (unsigned int)p_seskey);
+          } else if(!strcmp(p_fmmode, "new")){
             if(p_fmfilebuf && p_fmfilesiz > 0){
               const char *name = p_fmname;
               if(*name == '\0') name = p_fmfilename;
@@ -1299,6 +1338,7 @@ static void dosession(TCMPOOL *mpool){
   if(auth && p_user != '\0'){
     tcmapput2(vars, "username", p_user);
     if(userinfo && *userinfo != '\0') tcmapput2(vars, "userinfo", userinfo);
+    if(seskey > 0) tcmapprintf(vars, "seskey", "%llu", (unsigned long long)seskey);
   }
   if(auth) tcmapput2(vars, "auth", "true");
   if(admin) tcmapput2(vars, "admin", "true");
@@ -1317,6 +1357,7 @@ static void dosession(TCMPOOL *mpool){
   tcmapput2(vars, "scheme", p_scheme);
   tcmapput2(vars, "remotehost", p_remotehost);
   tcmapput2(vars, "hostname", p_hostname);
+  tcmapput2(vars, "referrer", p_referrer);
   tcmapput2(vars, "useragent", p_useragent);
   tcmapput2(vars, "userlang", p_userlang);
   const char *mtype = "text/html; charset=UTF-8";
